﻿using System;
using System.Collections.Generic;
using System.Text;
using Opc.Ua;
using System.Security.Cryptography.X509Certificates;
using Opc.Ua.Client;

namespace Goldammer.Helper.OPC {
    public class MAIOPCConnectionUA {

        #region private
       
        #region methods
      
        private static ApplicationConfiguration CreateClientConfiguration() {
            // The application configuration can be loaded from any file.
            // ApplicationConfiguration.Load() method loads configuration by looking up a file path in the App.config.
            // This approach allows applications to share configuration files and to update them.
            ApplicationConfiguration configuration = new ApplicationConfiguration();

            // Step 1 - Specify the server identity.
            configuration.ApplicationName = "UA Test Client";
            configuration.ApplicationType = ApplicationType.Client;
            configuration.ApplicationUri = "http://localhost/VendorId/ApplicationId/InstanceId";
            configuration.ProductUri = "http://VendorId/ProductId/VersionId";

            configuration.SecurityConfiguration = new SecurityConfiguration();

            // Step 2 - Specify the server's application instance certificate.

            // Application instance certificates must be placed in a windows certficate store because that is 
            // the best way to protect the private key. Certificates in a store are identified with 4 parameters:
            // StoreLocation, StoreName, SubjectName and Thumbprint.
            //
            // In this example the following values are used:
            // 
            //   LocalMachine    - use the machine wide certificate store.
            //   Personal        - use the store for individual certificates.
            //   ApplicationName - use the application name as a search key.   

            configuration.SecurityConfiguration.ApplicationCertificate = new CertificateIdentifier();
            configuration.SecurityConfiguration.ApplicationCertificate.StoreType = CertificateStoreType.Windows;
            configuration.SecurityConfiguration.ApplicationCertificate.StorePath = "LocalMachine\\My";
            configuration.SecurityConfiguration.ApplicationCertificate.SubjectName = configuration.ApplicationName;

            // trust all applications installed on the same machine.
            configuration.SecurityConfiguration.TrustedPeerCertificates.StoreType = CertificateStoreType.Windows;
            configuration.SecurityConfiguration.TrustedPeerCertificates.StorePath = "LocalMachine\\My";

            // find the certificate in the store.
            X509Certificate2 clientCertificate = configuration.SecurityConfiguration.ApplicationCertificate.Find(true);

            // create a new certificate if one not found.
            if (clientCertificate == null) {
                // this code would normally be called as part of the installer - called here to illustrate.
                // create a new certificate an place it in the LocalMachine/Personal store.
                clientCertificate = CertificateFactory.CreateCertificate(
                    configuration.SecurityConfiguration.ApplicationCertificate.StoreType,
                    configuration.SecurityConfiguration.ApplicationCertificate.StorePath,
                    configuration.ApplicationUri,
                    configuration.ApplicationName,
                    null,
                    null,
                    1024,
                    120);
            }

            // Step 3 - Specify the supported transport configurations.

            // The SDK requires Binding objects which are sub-types of Opc.Ua.Bindings.BaseBinding
            // These two lines add support for SOAP/HTTP w/ WS-* and UA-TCP. Support for other protocols 
            // such as .NET TCP can be added but they would not be considered interoperable across different vendors. 
            // Only one binding per URL scheme is allowed.
            configuration.TransportConfigurations.Add(new TransportConfiguration(Utils.UriSchemeOpcTcp, typeof(Opc.Ua.Bindings.UaTcpBinding)));
            configuration.TransportConfigurations.Add(new TransportConfiguration(Utils.UriSchemeHttp, typeof(Opc.Ua.Bindings.UaSoapXmlBinding)));

            // Step 4 - Specify the supported transport quotas.

            // The transport quotas are used to set limits on the contents of messages and are
            // used to protect against DOS attacks and rogue clients. They should be set to
            // reasonable values.
            configuration.TransportQuotas = new TransportQuotas();
            configuration.TransportQuotas.OperationTimeout = 360000;
            configuration.TransportQuotas.MaxStringLength = 67108864;
            configuration.ServerConfiguration = new ServerConfiguration();

            // Step 5 - Specify the client specific configuration.
            configuration.ClientConfiguration = new ClientConfiguration();
            configuration.ClientConfiguration.DefaultSessionTimeout = 360000;

            // Step 6 - Validate the configuration.

            // This step checks if the configuration is consistent and assigns a few internal variables
            // that are used by the SDK. This is called automatically if the configuration is loaded from
            // a file using the ApplicationConfiguration.Load() method.          
            configuration.Validate(ApplicationType.Client);

            return configuration;
        }

        private static EndpointDescription CreateEndpointDescription(string Url) {
            // create the endpoint description.
            EndpointDescription endpointDescription = new EndpointDescription();

            //endpointDescription.EndpointUrl = Utils.Format("opc.tcp://{0}:4841", System.Net.Dns.GetHostName());
            endpointDescription.EndpointUrl = Url;

            // specify the security policy to use.
            //endpointDescription.SecurityPolicyUri = SecurityPolicies.Basic128Rsa15;
            endpointDescription.SecurityPolicyUri = SecurityPolicies.None;
            //endpointDescription.SecurityMode      = MessageSecurityMode.SignAndEncrypt;
            endpointDescription.SecurityMode = MessageSecurityMode.None;

            // specify the transport profile.
            endpointDescription.TransportProfileUri = Profiles.WsHttpXmlOrBinaryTransport;

            // load the the server certificate from the local certificate store.
            CertificateIdentifier certificateIdentifier = new CertificateIdentifier();

            certificateIdentifier.StoreType = CertificateStoreType.Windows;
            certificateIdentifier.StorePath = "LocalMachine\\My";

            certificateIdentifier.SubjectName = "UA Test Client";

            X509Certificate2 serverCertificate = certificateIdentifier.Find();

            if (serverCertificate == null) {
                throw ServiceResultException.Create(StatusCodes.BadCertificateInvalid, "Could not find server certificate: {0}", certificateIdentifier.SubjectName);
            }

            endpointDescription.ServerCertificate = serverCertificate.GetRawCertData();

            return endpointDescription;
        }

        private static void GetTags(Browser browser, NodeId node, Dictionary<NodeId, Type> dict) {

            // Browse filter: only follow hierarchical references.
            browser.ReferenceTypeId = ReferenceTypes.HierarchicalReferences;

            ReferenceDescriptionCollection browseResults = null;
            try {
                // The actual call to the server.
                browseResults = browser.Browse(node);
                foreach (ReferenceDescription r in browseResults) {
                    if (r.DisplayName.Text.StartsWith("_"))
                        continue;
                    ReadValueIdCollection attribs = BuildAttributeList(r);
                    DataValueCollection results = null;
                    DiagnosticInfoCollection diagnosticInfos = null;
                    browser.Session.Read(null, 0, TimestampsToReturn.Both, attribs, out results, out diagnosticInfos);
                    Type myType = GetTypeOfLeaf(browser.Session,attribs, results);
                    if (myType != null) {
                        dict.Add((NodeId)r.NodeId, myType);
                    }
                    else {
                        GetTags(browser, (NodeId)r.NodeId, dict);
                    }

                }
            }
            catch (Exception e) {

                throw e;
            }
        }

        private static Type GetTypeOfLeaf(Session session, ReadValueIdCollection attribs, DataValueCollection results) {

            int i = 0;
            bool Writeable = false;
            string rer = null;
            foreach (ReadValueId att in attribs) {
                string handle = (string)att.Handle;
                if (handle.Equals("DataType")) {
                    string attributeValue = results[i].ToString();
                    NodeId node = (NodeId)results[i].Value;
                    INode datatype = session.NodeCache.Find(node);
                    if (datatype != null)
                        rer = datatype.ToString();
                }
                if (handle.Equals("UserAccessLevel")) {
                    object saa = results[i].Value;
                    byte levl = (byte)saa;
                    if ((levl & 2) == 2)
                        Writeable = true;
                }
                i++;
            }
            try {
                if (Writeable)
                    return Type.GetType("System." + rer);
            }
            catch {
            }
            return null;
        }

        private static ReadValueIdCollection BuildAttributeList(ReferenceDescription refDescription) {
            // Build list of attributes to read.
            ReadValueIdCollection nodesToRead = new ReadValueIdCollection();

            // Add default attributes (for all nodeclasses)
            AddAttribute((NodeId)refDescription.NodeId, Attributes.NodeId, nodesToRead);
            AddAttribute((NodeId)refDescription.NodeId, Attributes.NodeClass, nodesToRead);
            AddAttribute((NodeId)refDescription.NodeId, Attributes.BrowseName, nodesToRead);
            AddAttribute((NodeId)refDescription.NodeId, Attributes.DisplayName, nodesToRead);
            AddAttribute((NodeId)refDescription.NodeId, Attributes.Description, nodesToRead);
            AddAttribute((NodeId)refDescription.NodeId, Attributes.WriteMask, nodesToRead);
            AddAttribute((NodeId)refDescription.NodeId, Attributes.UserWriteMask, nodesToRead);

            // Add nodeclass specific attributes
            switch (refDescription.NodeClass) {
                case NodeClass.Object:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.EventNotifier, nodesToRead);
                    break;
                case NodeClass.Variable:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.Value, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.DataType, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.ValueRank, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.ArrayDimensions, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.AccessLevel, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.UserAccessLevel, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.MinimumSamplingInterval, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.Historizing, nodesToRead);
                    break;
                case NodeClass.Method:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.Executable, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.UserExecutable, nodesToRead);
                    break;
                case NodeClass.ObjectType:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.IsAbstract, nodesToRead);
                    break;
                case NodeClass.VariableType:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.Value, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.DataType, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.ValueRank, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.ArrayDimensions, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.IsAbstract, nodesToRead);
                    break;
                case NodeClass.ReferenceType:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.IsAbstract, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.Symmetric, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.InverseName, nodesToRead);
                    break;
                case NodeClass.DataType:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.IsAbstract, nodesToRead);
                    break;
                case NodeClass.View:
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.ContainsNoLoops, nodesToRead);
                    AddAttribute((NodeId)refDescription.NodeId, Attributes.EventNotifier, nodesToRead);
                    break;
                default:
                    break;
            }
            return nodesToRead;
        }

        private static void AddAttribute(NodeId node, uint attributeId, ReadValueIdCollection nodesToRead) {
            // Get NodeId from tree node .
            ReadValueId attributeToRead = new ReadValueId();
            attributeToRead.NodeId = node;
            attributeToRead.AttributeId = attributeId;
            attributeToRead.Handle = Attributes.GetBrowseName(attributeId);
            nodesToRead.Add(attributeToRead);
        }
      
        #endregion

        #region event Handlers

        private static void Session_KeepAlive(Session session, KeepAliveEventArgs e) {

        }
   
        private static void CertificateValidator_CertificateValidation(CertificateValidator validator, CertificateValidationEventArgs e) {
            // Accept all certificate -> better ask user
            e.Accept = true;
        }

        #endregion

        #endregion

        public static Session BuildSessionAndConnect(string Url) {
            try {
                // Create the configuration.     
                ApplicationConfiguration configuration = CreateClientConfiguration();

                // Create the endpoint description.
                EndpointDescription endpointDescription = CreateEndpointDescription(Url);

                // Create the endpoint configuration (use the application configuration to provide default values).
                EndpointConfiguration endpointConfiguration = EndpointConfiguration.Create(configuration);

                // The default timeout for a requests sent using the channel.
                endpointConfiguration.OperationTimeout = 300000;

                // Use the pure binary encoding on the wire.
                endpointConfiguration.UseBinaryEncoding = true;

                // Create the endpoint.
                ConfiguredEndpoint endpoint = new ConfiguredEndpoint(null, endpointDescription, endpointConfiguration);

                // Create the binding factory.
                BindingFactory bindingFactory = BindingFactory.Create(configuration);

                // Update endpoint description using the discovery endpoint.
                if (endpoint.UpdateBeforeConnect) {
                    endpoint.UpdateFromServer(bindingFactory);
                    endpointDescription = endpoint.Description;
                    endpointConfiguration = endpoint.Configuration;
                }

                X509Certificate2 clientCertificate = configuration.SecurityConfiguration.ApplicationCertificate.Find();

                // Set up a callback to handle certificate validation errors.

                configuration.CertificateValidator.CertificateValidation += new CertificateValidationEventHandler(CertificateValidator_CertificateValidation);

                // Initialize the channel which will be created with the server.
                SessionChannel channel = SessionChannel.Create(
                    configuration,
                    endpointDescription,
                    endpointConfiguration,
                    bindingFactory,
                    clientCertificate,
                    null);

                // Wrap the channel with the session object.
                // This call will fail if the server does not trust the client certificate.
                Session session = new Session(channel, configuration, endpoint);
                session.ReturnDiagnostics = DiagnosticsMasks.All;

                // Register keep alive callback.
                session.KeepAlive += new KeepAliveEventHandler(Session_KeepAlive);

                UserIdentity identity = new UserIdentity();

                // Create the session. This actually connects to the server.
                // Passing null for the user identity will create an anonymous session.
                session.Open("", identity);
                return session;
            }
            catch (Exception e) {
                throw e;
            }
        }

        public static Dictionary<NodeId, Type> BrowseSessionForWriteableTags(Session SessionToBrowse) {
            if(!SessionToBrowse.Connected)
                throw new Exception("Session is not connected !");
            Dictionary<NodeId, Type> ret = new Dictionary<NodeId, Type>();
            NodeId rootnode = new NodeId(Objects.RootFolder, 0);
            Browser browser = new Browser(SessionToBrowse);
            GetTags(browser, rootnode, ret);
            return ret;
        }

    }
}
