﻿Imports System.Text
Imports Opc.Ua
Imports System.Security.Cryptography.X509Certificates
Imports Opc.Ua.Client

Namespace Goldammer.Helper.OPC
	Public Class MAIOPCConnectionUA

		#Region "private"

		#Region "methods"

		Private Shared Function CreateClientConfiguration() As ApplicationConfiguration
			' The application configuration can be loaded from any file.
			' ApplicationConfiguration.Load() method loads configuration by looking up a file path in the App.config.
			' This approach allows applications to share configuration files and to update them.
			Dim configuration As New ApplicationConfiguration()

			' Step 1 - Specify the server identity.
			configuration.ApplicationName = "UA Test Client"
			configuration.ApplicationType = ApplicationType.Client
			configuration.ApplicationUri = "http://localhost/VendorId/ApplicationId/InstanceId"
			configuration.ProductUri = "http://VendorId/ProductId/VersionId"

			configuration.SecurityConfiguration = New SecurityConfiguration()

			' Step 2 - Specify the server's application instance certificate.

			' Application instance certificates must be placed in a windows certficate store because that is 
			' the best way to protect the private key. Certificates in a store are identified with 4 parameters:
			' StoreLocation, StoreName, SubjectName and Thumbprint.
			'
			' In this example the following values are used:
			' 
			'   LocalMachine    - use the machine wide certificate store.
			'   Personal        - use the store for individual certificates.
			'   ApplicationName - use the application name as a search key.   

			configuration.SecurityConfiguration.ApplicationCertificate = New CertificateIdentifier()
			configuration.SecurityConfiguration.ApplicationCertificate.StoreType = CertificateStoreType.Windows
			configuration.SecurityConfiguration.ApplicationCertificate.StorePath = "LocalMachine\My"
			configuration.SecurityConfiguration.ApplicationCertificate.SubjectName = configuration.ApplicationName

			' trust all applications installed on the same machine.
			configuration.SecurityConfiguration.TrustedPeerCertificates.StoreType = CertificateStoreType.Windows
			configuration.SecurityConfiguration.TrustedPeerCertificates.StorePath = "LocalMachine\My"

			' find the certificate in the store.
			Dim clientCertificate As X509Certificate2 = configuration.SecurityConfiguration.ApplicationCertificate.Find(True)

			' create a new certificate if one not found.
			If clientCertificate Is Nothing Then
				' this code would normally be called as part of the installer - called here to illustrate.
				' create a new certificate an place it in the LocalMachine/Personal store.
				clientCertificate = CertificateFactory.CreateCertificate(configuration.SecurityConfiguration.ApplicationCertificate.StoreType, configuration.SecurityConfiguration.ApplicationCertificate.StorePath, configuration.ApplicationUri, configuration.ApplicationName, Nothing, Nothing, 1024, 120)
			End If

			' Step 3 - Specify the supported transport configurations.

			' The SDK requires Binding objects which are sub-types of Opc.Ua.Bindings.BaseBinding
			' These two lines add support for SOAP/HTTP w/ WS-* and UA-TCP. Support for other protocols 
			' such as .NET TCP can be added but they would not be considered interoperable across different vendors. 
			' Only one binding per URL scheme is allowed.
            configuration.TransportConfigurations.Add(New TransportConfiguration(Utils.UriSchemeOpcTcp, GetType(Global.Opc.Ua.Bindings.UaTcpBinding)))
            configuration.TransportConfigurations.Add(New TransportConfiguration(Utils.UriSchemeHttp, GetType(Global.Opc.Ua.Bindings.UaSoapXmlBinding)))

			' Step 4 - Specify the supported transport quotas.

			' The transport quotas are used to set limits on the contents of messages and are
			' used to protect against DOS attacks and rogue clients. They should be set to
			' reasonable values.
			configuration.TransportQuotas = New TransportQuotas()
			configuration.TransportQuotas.OperationTimeout = 360000
			configuration.TransportQuotas.MaxStringLength = 67108864
			configuration.ServerConfiguration = New ServerConfiguration()

			' Step 5 - Specify the client specific configuration.
			configuration.ClientConfiguration = New ClientConfiguration()
			configuration.ClientConfiguration.DefaultSessionTimeout = 360000

			' Step 6 - Validate the configuration.

			' This step checks if the configuration is consistent and assigns a few internal variables
			' that are used by the SDK. This is called automatically if the configuration is loaded from
			' a file using the ApplicationConfiguration.Load() method.          
			configuration.Validate(ApplicationType.Client)

			Return configuration
		End Function

		Private Shared Function CreateEndpointDescription(ByVal Url As String) As EndpointDescription
			' create the endpoint description.
			Dim endpointDescription As New EndpointDescription()

			'endpointDescription.EndpointUrl = Utils.Format("opc.tcp://{0}:4841", System.Net.Dns.GetHostName());
			endpointDescription.EndpointUrl = Url

			' specify the security policy to use.
			'endpointDescription.SecurityPolicyUri = SecurityPolicies.Basic128Rsa15;
			endpointDescription.SecurityPolicyUri = SecurityPolicies.None
			'endpointDescription.SecurityMode      = MessageSecurityMode.SignAndEncrypt;
			endpointDescription.SecurityMode = MessageSecurityMode.None

			' specify the transport profile.
			endpointDescription.TransportProfileUri = Profiles.WsHttpXmlOrBinaryTransport

			' load the the server certificate from the local certificate store.
			Dim certificateIdentifier As New CertificateIdentifier()

			certificateIdentifier.StoreType = CertificateStoreType.Windows
			certificateIdentifier.StorePath = "LocalMachine\My"

			certificateIdentifier.SubjectName = "UA Test Client"

			Dim serverCertificate As X509Certificate2 = certificateIdentifier.Find()

			If serverCertificate Is Nothing Then
				Throw ServiceResultException.Create(StatusCodes.BadCertificateInvalid, "Could not find server certificate: {0}", certificateIdentifier.SubjectName)
			End If

			endpointDescription.ServerCertificate = serverCertificate.GetRawCertData()

			Return endpointDescription
		End Function

		Private Shared Sub GetTags(ByVal browser As Browser, ByVal node As NodeId, ByVal dict As Dictionary(Of NodeId, Type))

			' Browse filter: only follow hierarchical references.
			browser.ReferenceTypeId = ReferenceTypes.HierarchicalReferences

			Dim browseResults As ReferenceDescriptionCollection = Nothing
			Try
				' The actual call to the server.
				browseResults = browser.Browse(node)
				For Each r As ReferenceDescription In browseResults
					If r.DisplayName.Text.StartsWith("_") Then
						Continue For
					End If
					Dim attribs As ReadValueIdCollection = BuildAttributeList(r)
					Dim results As DataValueCollection = Nothing
					Dim diagnosticInfos As DiagnosticInfoCollection = Nothing
					browser.Session.Read(Nothing, 0, TimestampsToReturn.Both, attribs, results, diagnosticInfos)
					Dim myType As Type = GetTypeOfLeaf(browser.Session,attribs, results)
					If myType IsNot Nothing Then
						dict.Add(CType(r.NodeId, NodeId), myType)
					Else
						GetTags(browser, CType(r.NodeId, NodeId), dict)
					End If

				Next r
			Catch e As Exception

				Throw e
			End Try
		End Sub

		Private Shared Function GetTypeOfLeaf(ByVal session As Session, ByVal attribs As ReadValueIdCollection, ByVal results As DataValueCollection) As Type

			Dim i As Integer = 0
			Dim Writeable As Boolean = False
			Dim rer As String = Nothing
			For Each att As ReadValueId In attribs
				Dim handle As String = CStr(att.Handle)
				If handle.Equals("DataType") Then
					Dim attributeValue As String = results(i).ToString()
					Dim node As NodeId = CType(results(i).Value, NodeId)
					Dim datatype As INode = session.NodeCache.Find(node)
					If datatype IsNot Nothing Then
						rer = CType(datatype, Object).ToString()
					End If
				End If
				If handle.Equals("UserAccessLevel") Then
					Dim saa As Object = results(i).Value
					Dim levl As Byte = CByte(saa)
					If (levl And 2) = 2 Then
						Writeable = True
					End If
				End If
				i += 1
			Next att
			Try
				If Writeable Then
					Return Type.GetType("System." & rer)
				End If
			Catch
			End Try
			Return Nothing
		End Function

		Private Shared Function BuildAttributeList(ByVal refDescription As ReferenceDescription) As ReadValueIdCollection
			' Build list of attributes to read.
			Dim nodesToRead As New ReadValueIdCollection()

			' Add default attributes (for all nodeclasses)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.NodeId, nodesToRead)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.NodeClass, nodesToRead)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.BrowseName, nodesToRead)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.DisplayName, nodesToRead)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.Description, nodesToRead)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.WriteMask, nodesToRead)
			AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.UserWriteMask, nodesToRead)

			' Add nodeclass specific attributes
			Select Case refDescription.NodeClass
				Case NodeClass.Object
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.EventNotifier, nodesToRead)
				Case NodeClass.Variable
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.Value, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.DataType, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.ValueRank, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.ArrayDimensions, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.AccessLevel, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.UserAccessLevel, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.MinimumSamplingInterval, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.Historizing, nodesToRead)
				Case NodeClass.Method
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.Executable, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.UserExecutable, nodesToRead)
				Case NodeClass.ObjectType
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.IsAbstract, nodesToRead)
				Case NodeClass.VariableType
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.Value, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.DataType, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.ValueRank, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.ArrayDimensions, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.IsAbstract, nodesToRead)
				Case NodeClass.ReferenceType
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.IsAbstract, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.Symmetric, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.InverseName, nodesToRead)
				Case NodeClass.DataType
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.IsAbstract, nodesToRead)
				Case NodeClass.View
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.ContainsNoLoops, nodesToRead)
					AddAttribute(CType(refDescription.NodeId, NodeId), Attributes.EventNotifier, nodesToRead)
				Case Else
			End Select
			Return nodesToRead
		End Function

		Private Shared Sub AddAttribute(ByVal node As NodeId, ByVal attributeId As UInteger, ByVal nodesToRead As ReadValueIdCollection)
			' Get NodeId from tree node .
			Dim attributeToRead As New ReadValueId()
			attributeToRead.NodeId = node
			attributeToRead.AttributeId = attributeId
			attributeToRead.Handle = Attributes.GetBrowseName(attributeId)
			nodesToRead.Add(attributeToRead)
		End Sub

		#End Region

		#Region "event Handlers"

		Private Shared Sub Session_KeepAlive(ByVal session As Session, ByVal e As KeepAliveEventArgs)

		End Sub

		Private Shared Sub CertificateValidator_CertificateValidation(ByVal validator As CertificateValidator, ByVal e As CertificateValidationEventArgs)
			' Accept all certificate -> better ask user
			e.Accept = True
		End Sub

		#End Region

		#End Region

		Public Shared Function BuildSessionAndConnect(ByVal Url As String) As Session
			Try
				' Create the configuration.     
				Dim configuration As ApplicationConfiguration = CreateClientConfiguration()

				' Create the endpoint description.
				Dim endpointDescription As EndpointDescription = CreateEndpointDescription(Url)

				' Create the endpoint configuration (use the application configuration to provide default values).
				Dim endpointConfiguration As EndpointConfiguration = EndpointConfiguration.Create(configuration)

				' The default timeout for a requests sent using the channel.
				endpointConfiguration.OperationTimeout = 300000

				' Use the pure binary encoding on the wire.
				endpointConfiguration.UseBinaryEncoding = True

				' Create the endpoint.
				Dim endpoint As New ConfiguredEndpoint(Nothing, endpointDescription, endpointConfiguration)

				' Create the binding factory.
				Dim bindingFactory As BindingFactory = BindingFactory.Create(configuration)

				' Update endpoint description using the discovery endpoint.
				If endpoint.UpdateBeforeConnect Then
					endpoint.UpdateFromServer(bindingFactory)
					endpointDescription = endpoint.Description
					endpointConfiguration = endpoint.Configuration
				End If

				Dim clientCertificate As X509Certificate2 = configuration.SecurityConfiguration.ApplicationCertificate.Find()

				' Set up a callback to handle certificate validation errors.

				AddHandler configuration.CertificateValidator.CertificateValidation, AddressOf CertificateValidator_CertificateValidation

				' Initialize the channel which will be created with the server.
				Dim channel As SessionChannel = SessionChannel.Create(configuration, endpointDescription, endpointConfiguration, bindingFactory, clientCertificate, Nothing)

				' Wrap the channel with the session object.
				' This call will fail if the server does not trust the client certificate.
				Dim session As New Session(channel, configuration, endpoint)
				session.ReturnDiagnostics = DiagnosticsMasks.All

				' Register keep alive callback.
				AddHandler session.KeepAlive, AddressOf Session_KeepAlive

				Dim identity As New UserIdentity()

				' Create the session. This actually connects to the server.
				' Passing null for the user identity will create an anonymous session.
				session.Open("", identity)
				Return session
			Catch e As Exception
				Throw e
			End Try
		End Function

		Public Shared Function BrowseSessionForWriteableTags(ByVal SessionToBrowse As Session) As Dictionary(Of NodeId, Type)
			If Not SessionToBrowse.Connected Then
				Throw New Exception("Session is not connected !")
			End If
			Dim ret As New Dictionary(Of NodeId, Type)()
			Dim rootnode As New NodeId(Objects.RootFolder, 0)
			Dim browser As New Browser(SessionToBrowse)
			GetTags(browser, rootnode, ret)
			Return ret
		End Function

	End Class
End Namespace
